/**
 * Copyright Copyright 2010 Simon Andrews
 *
 *    This file is part of Difference Tracker.
 *
 *    Difference Tracker is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    Difference Tracker is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Difference Tracker; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.DifferenceTracker;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;

import uk.ac.bbsrc.babraham.DifferenceTracker.Config.DifferenceTrackerPreferences;
import uk.ac.bbsrc.babraham.DifferenceTracker.Config.EditDiffFilterPrefs;
import uk.ac.bbsrc.babraham.DifferenceTracker.Results.DiffFilterResultsTableModel;
import uk.ac.bbsrc.babraham.DifferenceTracker.Results.DifferenceResults;
import uk.ac.bbsrc.babraham.DifferenceTracker.Results.ResultsTable;

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.Macro;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;


public class Difference_Filter implements PlugInFilter {

	public static final String VERSION = "0.1";
	private ImagePlus imp = null;
	private boolean unattended = false;
	private DifferenceResults results;

	public void run(ImageProcessor ip) {

		System.out.println("File: "+imp.getTitle());
		
		if (! unattended) {
			EditDiffFilterPrefs d = new EditDiffFilterPrefs();
			if (d.wasCancelled()) return;
		}
		
				
		DifferenceTrackerPreferences prefs = DifferenceTrackerPreferences.getInstance();
		int offsetSize = prefs.differenceFrameOffset();
		int differenceThreshold = prefs.minAbsoluteDiff();
		
		results = new DifferenceResults(imp);
		
		int width = ip.getWidth();
		int height = ip.getHeight();


		for (int slice=imp.getImageStackSize();slice>offsetSize;slice--) {

			IJ.showStatus("Processing slice "+slice);
			ImageProcessor thisProcessor = imp.getImageStack().getProcessor(slice);
			ImageProcessor lastProcessor = imp.getImageStack().getProcessor(slice-offsetSize);

			imp.setSlice(slice);
			int [][] newIntensities = new int [width][height];


			// Work our the average intensity for everything which is moving
			Vector<Integer> movingIntensities = new Vector<Integer>();

			for (int w=0;w<width;w++) {
				for (int h=0;h<height;h++) {
					int thisPixel = thisProcessor.getPixel(w, h);

					int [] comparePixels = surroundingPixels(lastProcessor,w,h);

					int smallestDiff = 0;
					// Find the smallest diff
					for (int i=0;i<comparePixels.length;i++) {

						// We use -1 to denote a blank in the set of pixel values
						// these always occur after all the real data so we can
						// stop if we see one.
						if (comparePixels[i]<0) break;

						// We only want pixels which are brighter in this frame than the
						// compared frame otherwise we get ghosting.

						int diff = thisPixel-comparePixels[i];

						if (i == 0 || diff < smallestDiff) {
							smallestDiff = diff;
						}
												
					}

					if (smallestDiff < differenceThreshold) {
						smallestDiff = 0; // Set a threshold and remove negative diffs
					}
					else {
						// This is a moving pixel
						movingIntensities.add(thisPixel);
					}

					// Make the output nonlinear to emphasise the most changing points.
					newIntensities[w][h]=(int)Math.pow(smallestDiff, 2);
				}

			}

			Integer [] movingIntensityValues = movingIntensities.toArray(new Integer [0]);

			// The moving reference is the median intensity of the moving pixels.  We set
			// this to zero intially in case we don't have any moving pixels.
			double movingReference = 0;

			if (movingIntensityValues.length > 0) {

				// Work out the median intensity of the moving pixels.  We'll use this
				// as as sort of background value and calculate all future intensities
				// and counts using this as a reference.

				// The -1 is in case we have all pixels with the same intensity.  In this
				// case we want the cutoff to fall just below this value otherwise we end
				// up with counts of 0.
				movingReference = SimpleStats.median(movingIntensityValues)-1;
			}
			
			else {
				System.err.println("No moving pixels recorded");
			}



			// Work out the average moving intensity from the moving intensities
			// higher than the reference.
			double movingIntensity = 0;
			int movingCount = 0;
			for (int i=0;i<movingIntensityValues.length;i++) {
				if (movingIntensityValues[i]>=movingReference) {
					++movingCount;
					movingIntensity += movingIntensityValues[i]-movingReference;
				}
			}



			/*
			 * Now we use the moving reference intensity to go back through the data.
			 * 
			 * We're working out the total number of pixels which have intensities
			 * higher than the moving count.  We also work out the total intensity
			 * of these pixels
			 */

			/*
			 * Since we're going to encounter all of the moving pixels again we'll
			 * subtract these from the total before we start so we only get data
			 * for the static part of the image.
			 */

			double totalIntensity = 0;
			int aboveIntensityCount = 0;

			for (int w=0;w<width;w++) {
				for (int h=0;h<height;h++) {
					if (thisProcessor.getPixel(w, h) >= movingReference) {
						++aboveIntensityCount;
						totalIntensity += thisProcessor.getPixel(w, h)-movingReference;
					}
					thisProcessor.putPixel(w,h,newIntensities[w][h]);
				}
			}


			results.setMovingCount(movingCount, slice);
			results.setStaticCount(aboveIntensityCount-movingCount, slice);
			results.setMovingIntensity(movingIntensity, slice);
			results.setStaticIntensity(totalIntensity-movingIntensity,slice);

		}


		// We now need to delete the unused frames to stop them from interfering from
		// future calculations.

		for (int slice=offsetSize;slice>0;slice--) {
			IJ.showStatus("Deleting slice "+slice);
			
			/*
			 * This is somewhat convoluted.  You need to extract the stack, delete a slice
			 * and then add it back rather than just manipulating it directly, otherwise the
			 * wrong slice is removed.
			 */
			
			ImageStack stack =imp.getImageStack();
			stack.deleteSlice(1);
			imp.setStack(null, stack);
		}

		DiffFilterResultsTableModel resultsModel = new DiffFilterResultsTableModel(imp.getOriginalFileInfo().fileName,results);

		File exportFile = new File(imp.getOriginalFileInfo().directory+"difference_filter_output.txt");

		boolean headersWritten = exportFile.exists();

		PrintWriter pr = null;
		try {
			pr = new PrintWriter(new FileWriter(exportFile,true));
		} 
		catch (IOException e) {
			e.printStackTrace();
		}

		
		// We output to the file using the table model.  The model is 2 columns where the
		// first column are the headers and the second column are the values
		
		if (! headersWritten) {
			for (int i=0;i<resultsModel.getRowCount();i++) {
				pr.print(resultsModel.getValueAt(i, 0)); // Column 0 means the field names
				if (i == resultsModel.getRowCount()-1) {
					pr.print("\n");
				}
				else {
					pr.print("\t");
				}
			}			
		}

		// Now we can print out the actual data
		for (int i=0;i<resultsModel.getRowCount();i++) {
			pr.print(resultsModel.getValueAt(i, 1)); // Column 1 means the data
			if (i == resultsModel.getRowCount()-1) {
				pr.print("\n");
			}
			else {
				pr.print("\t");
			}
		}			
		
		pr.close();
		
		if (!unattended) {
			new ResultsTable(resultsModel);
		}


	}

	private int [] surroundingPixels (ImageProcessor p, int w, int h) {

		int [] values = new int[] {-1,-1,-1,-1,-1,-1,-1,-1,-1};

		int currentPos = 0;

		for (int x=w-1;x<=w+1;x++) {
			if (x<0 || x>=p.getWidth()) continue;
			for (int y=h-1;y<=h+1;y++) {
				if (y<0 || y>=p.getHeight()) continue;
				values[currentPos] = p.getPixel(x, y);
				++currentPos;
			}
		}

		return values;
	}

	public int setup(String args, ImagePlus imp) {
			
		// If they've not passed any arguments then we put up some options.
		// This allows us to run silently inside a macro without pausing
		// to display results.
		if (Macro.getOptions() == null) {
			unattended = false;
		}
		else {
			unattended = true;
		}

		
		this.imp = imp;
		return DOES_8G;
	}


	public static void main (String [] args) {
		new Difference_Filter();
	}
}

