/**
 * Copyright Copyright 2010 Simon Andrews
 *
 *    This file is part of Difference Tracker.
 *
 *    Difference Tracker is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    Difference Tracker is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Difference Tracker; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.DifferenceTracker;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Vector;

import uk.ac.bbsrc.babraham.DifferenceTracker.Config.DifferenceTrackerPreferences;


public class TrackedParticleCollection {

	/**
	 * This class manages a collection of tracked particles.  It allows the
	 * addition of new ParticleCollections and the integration of these with
	 * previously tracked particles.
	 * 
	 */
	
	private double initialFlexibility;
	private double subsequentFlexibility;
	
	private Vector<TrackedParticle> currentlyTrackedParticles = new Vector<TrackedParticle>();
	private Vector<TrackedParticle> previouslyTrackedParticles = new Vector<TrackedParticle>();
	
	public TrackedParticleCollection (ParticleCollection pc, int slice) {
		
		initialFlexibility = DifferenceTrackerPreferences.getInstance().initalFlexibility();
		subsequentFlexibility = DifferenceTrackerPreferences.getInstance().subsequentFlexibility();
		
		Particle [] particles = pc.getAvailableParticles(DifferenceTrackerPreferences.getInstance().minFeatureSize());
		
		for (int p=0;p<particles.length;p++) {
			currentlyTrackedParticles.add(new TrackedParticle(particles[p],slice));
		}
	}
	
	public void extendTrackedParticles (ParticleCollection pc, int slice) {

		Particle [] initialParticles = pc.getAvailableParticles(DifferenceTrackerPreferences.getInstance().minFeatureSize());
		Vector<Particle> particles = new Vector<Particle>();
		for (int i=0;i<initialParticles.length;i++) {
			particles.add(initialParticles[i]);
		}
		
		// We first go through any established tracks letting them select
		// the best particle from the choices available.  These particles
		// can be assigned to more than one track, but will then be
		// unavailable
		
		
		
		Vector<TrackedParticle> savedTrackedParticles = new Vector<TrackedParticle>();
		Vector<TrackedParticle> finishedTrackedParticles = new Vector<TrackedParticle>();

		HashSet<Particle> particlesToRemove = new HashSet<Particle>();
		
		Enumeration<TrackedParticle> t = currentlyTrackedParticles.elements();

		while (t.hasMoreElements()) {

			TrackedParticle tr = t.nextElement();
			
			// We only want particles with an established track
			if (tr.length()<2) continue;
			
			Particle bestParticle = null;
			double closestDiff = -1;
			
			Enumeration<Particle> p = particles.elements();

			while (p.hasMoreElements()) {

				Particle pa = p.nextElement();
				
				if (particlesToRemove.contains(pa)) continue;
				
				double distance = tr.wantsParticle(pa);
				if (distance > subsequentFlexibility) continue; // We're not interested at all

				if (bestParticle == null || distance < closestDiff) {
					bestParticle = pa;
					closestDiff = distance;
				}
			}
			
			if (bestParticle != null) {
//				System.out.println("Selecting particle at "+bestParticle.getCentrePosition().x+","+bestParticle.getCentrePosition().y+" as best match for "+tr.prediction().x+","+tr.prediction().y);
				tr.takeParticle(bestParticle);
				savedTrackedParticles.add(tr);
				particlesToRemove.add(bestParticle);
			}
			else {
				// If we get here then this tracked particle is finished
				finishedTrackedParticles.add(tr);
			}
		}
				
		// Now we do the same thing for the particles with only one
		// previous position
		
		t = currentlyTrackedParticles.elements();
		
		while (t.hasMoreElements()) {

			TrackedParticle tr = t.nextElement();
			
			// We only want particles without an established track
			if (tr.length() != 1) continue;
			
			Particle bestParticle = null;
			double closestDiff = -1;
			
			Enumeration<Particle> p = particles.elements();

			while (p.hasMoreElements()) {

				Particle pa = p.nextElement();
				
				if (particlesToRemove.contains(pa)) continue;
				
				double distance = tr.wantsParticle(pa);
				if (distance > initialFlexibility) continue; // We're not interested at all

				if (bestParticle == null || distance < closestDiff) {
					bestParticle = pa;
					closestDiff = distance;
				}
			}
			
			if (bestParticle != null) {
				tr.takeParticle(bestParticle);
				savedTrackedParticles.add(tr);
				particlesToRemove.add(bestParticle);
			}
			else {
				// If we get here then this tracked particle is finished
				finishedTrackedParticles.add(tr);
			}
		}

		
		currentlyTrackedParticles = savedTrackedParticles;
		
		Enumeration<TrackedParticle> e = finishedTrackedParticles.elements();
		while (e.hasMoreElements()) {
			previouslyTrackedParticles.add(e.nextElement());
		}
		
		// Any remaining particles get added as newly tracked particles
		Enumeration<Particle> p = particles.elements();
		while (p.hasMoreElements()) {
			Particle particle = p.nextElement();
			if (particlesToRemove.contains(particle)) continue;
			currentlyTrackedParticles.add(new TrackedParticle(particle,slice));
		}

	}

	public TrackedParticle [] getAllTrackedParticles (int minDuration) {
		return getAllTrackedParticles(minDuration, 0);
	}
	
	public TrackedParticle [] getAllTrackedParticles (int minDuration, int directionOfMovement) {
		Vector<TrackedParticle> keepers = new Vector<TrackedParticle>();
		
		Enumeration<TrackedParticle> current = currentlyTrackedParticles.elements();
		
//		System.out.println("Finding particles with direction "+directionOfMovement);
		
		while (current.hasMoreElements()) {
			TrackedParticle p = current.nextElement();
			if (directionOfMovement != 0 && p.getDirectionOfMovement() != directionOfMovement) continue;
			if (p.length() >= minDuration) {
				keepers.add(p);
			}
		}
		
		Enumeration<TrackedParticle> old = previouslyTrackedParticles.elements();
		
		while (old.hasMoreElements()) {
			TrackedParticle p = old.nextElement();
			if (directionOfMovement != 0 && p.getDirectionOfMovement() != directionOfMovement) continue;
			if (p.length() >= minDuration) {
				keepers.add(p);
			}
		}
		
		return keepers.toArray(new TrackedParticle[0]);
	}
	
	public ParticleSetStats SummariseMovement (int minDuration, int directionOfMovement, int stackSize) {
		
		/**
		 * Works out the average speed and average number of tracks per
		 * frame
		 */

		ParticleSetStats stats = new ParticleSetStats();

		TrackedParticle [] particles = getAllTrackedParticles(minDuration, directionOfMovement);
				
		System.out.println("Total tracked particles in direction "+directionOfMovement+" is "+particles.length);
		
		stats.totalCount = particles.length;
		
		int [] trackLengths = new int [particles.length];
		for (int i=0;i<particles.length;i++) {
			trackLengths[i] = particles[i].length();
		}
		stats.averageTrackDuration = SimpleStats.mean(trackLengths);
		
		double [] volumes = new double [particles.length];
		for (int i=0;i<particles.length;i++) {
			volumes[i] = particles[i].averageVolume();
		}
		
		stats.averageVolume = SimpleStats.mean(volumes);
		
		double [] maxSpeeds = new double [particles.length];
		for (int i=0;i<particles.length;i++) {
			maxSpeeds[i] = particles[i].maxSpeed();
		}
		
		stats.averageMaxSpeed = SimpleStats.mean(maxSpeeds);
		
		double totalDistance = 0;
		int totalFrameCounts = 0;
		
		for (int p=0;p<particles.length;p++) {
			totalDistance += particles[p].distanceMoved();
			totalFrameCounts += particles[p].length()-1;
		}
		
		stats.averageSpeed = totalDistance/totalFrameCounts;
		
		int [] counts = new int [stackSize];
		for (int slice = 1; slice <=stackSize;slice++) {
			for (int p=0;p<particles.length;p++) {
				if (particles[p].presentInSlice(slice)) {
					counts[slice-1]++;
				}
			}
		}
		
		stats.averageCount = SimpleStats.mean(counts);
		
		return stats;
	}
	
	
}
